#include "_Plugin_Helper.h"
#ifdef USES_P220
//#######################################################################################################
//#################################### Plugin 220: PID ##################################################
//#######################################################################################################

#include "src/PluginStructs/P220_data_struct.h"

#define PLUGIN_220
#define PLUGIN_ID_220         220
#define PLUGIN_NAME_220       "Regulator - PID"
#define PLUGIN_VALUENAME1_220 "PID"
#define PLUGIN_VALUENAME2_220 "TimeDiff"
#define PLUGIN_VALUENAME3_220 "Input"
#define PLUGIN_VALUENAME4_220 "LastInput"

boolean Plugin_220(byte function, struct EventStruct *event, String& string)
{
  boolean success = false;

  switch (function)
  {
    case PLUGIN_DEVICE_ADD:
      {
        Device[++deviceCount].Number              = PLUGIN_ID_220;
        Device[deviceCount].Type                  = DEVICE_TYPE_DUMMY;
        Device[deviceCount].VType                 = Sensor_VType::SENSOR_TYPE_SINGLE;
        Device[deviceCount].Ports                 = 0;
        Device[deviceCount].PullUpOption          = false;
        Device[deviceCount].InverseLogicOption    = false;
        Device[deviceCount].FormulaOption         = false;
        Device[deviceCount].DecimalsOnly          = true;
        Device[deviceCount].ValueCount            = 4;
        Device[deviceCount].SendDataOption        = true;
        Device[deviceCount].TimerOption           = true;
        Device[deviceCount].GlobalSyncOption      = true;
        //Device[deviceCount].OutputDataType        = Output_Data_type_t::All;
        break;
      }

    case PLUGIN_GET_DEVICENAME:
      {
        string = F(PLUGIN_NAME_220);
        break;
      }

    case PLUGIN_GET_DEVICEVALUENAMES:
      {
        strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[0], PSTR(PLUGIN_VALUENAME1_220));
        strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[1], PSTR(PLUGIN_VALUENAME2_220));
        strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[2], PSTR(PLUGIN_VALUENAME3_220));
        strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[3], PSTR(PLUGIN_VALUENAME4_220));
        break;
      }

    case PLUGIN_WEBFORM_LOAD:
      {
        addHtml(F("<TR><TD>Check Task:<TD>"));
        addTaskSelect(F("p220_task"), PCONFIG(0));

        LoadTaskSettings(PCONFIG(0)); // we need to load the values from another task for selection!
        addHtml(F("<TR><TD>Check Value:<TD>"));
        addTaskValueSelect(F("p220_value"), PCONFIG(1), PCONFIG(0));

        addFormCheckBox(F("Enable Proportional on Measurement"), F("p220_pOnM"), PCONFIG(2));        
      	addFormTextBox(F("SetPoint"), F("p220_setvalue"), String(PCONFIG_FLOAT(0)), 8);
        addFormTextBox(F("MaxUnder"), F("p220_maxunder"), String(PCONFIG(3)), 8);
        addFormTextBox(F("kP"), F("p220_kp"), String(PCONFIG_FLOAT(1)), 8);
        addFormTextBox(F("kI"), F("p220_ki"), String(PCONFIG_FLOAT(2)), 8);
        addFormTextBox(F("kD"), F("p220_kd"), String(PCONFIG_FLOAT(3)), 8);
        addFormTextBox(F("minOut"), F("p220_minout"), String(PCONFIG(4)), 8);
        addFormTextBox(F("maxOut"), F("p220_maxout"), String(PCONFIG(5)), 8);

        //LoadTaskSettings(event->TaskIndex); // we need to restore our original taskvalues!
        success = true;
        break;
      }

    case PLUGIN_WEBFORM_SAVE:
      {
        PCONFIG(0) = getFormItemInt(F("p220_task"));
        PCONFIG(1) = getFormItemInt(F("p220_value"));
        PCONFIG(2) = isFormItemChecked(F("p220_pOnM"));
        PCONFIG(3) = getFormItemInt(F("p220_maxunder"));
        PCONFIG(4) = getFormItemInt(F("p220_minout"));
        PCONFIG(5) = getFormItemInt(F("p220_maxout"));
        
        PCONFIG_FLOAT(0) = getFormItemFloat(F("p220_setvalue"));
        PCONFIG_FLOAT(1) = getFormItemFloat(F("p220_kp"));
        PCONFIG_FLOAT(2) = getFormItemFloat(F("p220_ki"));
        PCONFIG_FLOAT(3) = getFormItemFloat(F("p220_kd"));
        
        success = true;
        break;
      }

    case PLUGIN_INIT:
    {
        initPluginTaskData(event->TaskIndex, new (std::nothrow) P220_data_struct(millis(), 0, 0));
        P220_data_struct *P220_data = static_cast<P220_data_struct *>(getPluginTaskData(event->TaskIndex));

        success = true;
        break;
    }

    case PLUGIN_SET_CONFIG:
      {
        String command = parseString(string, 1);
        if (command == F("PID"))
        {
          String value = parseString(string, 2);
          double result=0;
          if (!isError(Calculate(value, result))) {
            PCONFIG_FLOAT(0) = result;
            SaveSettings();
            success = true;
          }
        }
        break;
      }

    case PLUGIN_GET_CONFIG:
      {
        String command = parseString(string, 1);
        if (command == F("PID"))
        {
          string = PCONFIG_FLOAT(0);
          success = true;
        }
        break;
      }

    case PLUGIN_SET_DEFAULTS:
    {
      PCONFIG_FLOAT(0) = 50.0;
      PCONFIG_FLOAT(1) = 9.1;
      PCONFIG_FLOAT(2) = 0.3;
      PCONFIG_FLOAT(3) = 1.8;
      
      PCONFIG(3) = 40;
      PCONFIG(4) = 0;
      PCONFIG(5) = 1000;
      
      success = true;
      break;
    }
    
    case PLUGIN_ONCE_A_SECOND:
      {
        P220_data_struct *P220_data = static_cast<P220_data_struct *>(getPluginTaskData(event->TaskIndex));
        // we're checking a var from another task, so calculate that basevar
        taskIndex_t TaskIndex = PCONFIG(0);
        byte BaseVarIndex = TaskIndex * VARS_PER_TASK + PCONFIG(1);
       
        // load var
        float SetPoint = PCONFIG_FLOAT(0);
        int maxUnder = PCONFIG(3);
        float kP = PCONFIG_FLOAT(1);
        float kI = PCONFIG_FLOAT(2);
        float kD = PCONFIG_FLOAT(3);
        int minOut = PCONFIG(4);
        int maxOut = PCONFIG(5);       
        
        float Input = UserVar[BaseVarIndex]; // SetPoint;
        float lastInput = P220_data->get_lastInput();
        if(Input < -100)
           Input = lastInput;
        float Error = SetPoint - Input;
        float dInput = Input - lastInput;

        float outputSum = P220_data->get_outputSum();
        outputSum += (kI * Error);

        boolean pOnE = PCONFIG(2);
        
        if(!pOnE) outputSum -= kP * dInput;

        if(outputSum > maxOut) outputSum = maxOut;
        else if(outputSum < minOut) outputSum = minOut;

        float PID = 0;
        if(pOnE) PID += kP * Error;

        PID += outputSum - kD * dInput;

        if(PID > maxOut) PID = maxOut;
        else if(PID < minOut) PID = minOut;

        if(Input < maxUnder)
           PID = maxOut;

        unsigned long TimeNow = millis();
        unsigned long TimeLast = P220_data->get_lastTime();
        unsigned long TimeDiff = TimeNow - TimeLast;

        P220_data->save_parameters(TimeNow, Input, outputSum);

        UserVar[event->BaseVarIndex] = PID;
        UserVar[event->BaseVarIndex + 1] = TimeDiff;
        UserVar[event->BaseVarIndex + 2] = Input;
        UserVar[event->BaseVarIndex + 3] = lastInput;    

        //String log = F("PID: PID ");
        //log += PID;
        //log = F("PID: TimeDiff ");
        //log += TimeDiff;
        //log = F("PID: Input ");
        //log += Input;
        //log = F("PID: lastInput ");
        //log += lastInput;
        //addLog(LOG_LEVEL_INFO, log);
        sendData(event);
        
        success = true;
        break;
      }

  }
  return success;
}
#endif // USES_P220
